#include "aldor.as"
#include "aldorio.as"

Edge(T: PrimitiveType): PrimitiveType with {
      edge: (T, T) -> %;
      p1: % -> T;
      p0: % -> T;

       if T has OutputType then OutputType;
} == add {
  Rep == Record(p0: T, p1: T);
  import from Rep;

  (a: %) = (b: %): Boolean == {
      p0 a = p0 b and p1 a = p1 b;
  }

  edge(s: T, e: T): % == per [s, e];
  p0(e: %): T == (rep e).p0;
  p1(e: %): T == (rep e).p1;

  if T has OutputType then {
     (o: TextWriter) << (e: %): TextWriter == { 
       import from String; 
       import from T;
       o << p0 e << " -> " << p1 e;
     }
   }

}

-- It would be nice to have a generic 'Set' type...
define B__GraphCategory(T: PrimitiveType): Category == with {
       nodes: % -> Set T;
       edges: % -> Set Edge T;

       outEdges: (%, T) -> List Edge T;
       inEdges:  (%, T) -> List Edge T;
       node?: (T, %) -> Boolean;

       if T has OutputType then OutputType;
}

B__Graph(T: Join(HashType, OutputType)): B__GraphCategory T with {
    graph:    List T -> %;
    graph: Generator T -> %;

    addLink!: (%, T, T) -> ();
    sortedNodes: % -> Set T;    
    ancestors: (%, T) -> Set T;

    #: % -> MachineInteger;
} == add {

  NdInfo == Record(inEdges: List Edge T, outEdges: List Edge T);
  Rep ==> HashTable(T, NdInfo);

  import from Rep;
  import from NdInfo;

  local emptyNd(): NdInfo == [ [], [] ];

  #(g: %): MachineInteger == numberOfEntries rep(g);

  local bracket(g: Generator Cross(T, NdInfo)): Rep == {
      tbl: Rep := table();
      for pair in g repeat { (n, v) := pair; tbl.n := v;}
      tbl
  }

  node?(node: T, g: %) : Boolean == {
	import from Partial NdInfo;
	not failed? find(node, rep(g))
  }


  graph(nodes: List T): % == {
      per [ (node, [[], []]) for node in nodes];
  } 

  graph(nodes: Generator T): % == {
      per [ (node, [[], []]) for node in nodes];
  } 

  nodes(g: %): Set T == [keys rep g];
  
  edges(g: %): Set Edge T == {
     [generate { for ndinfo in entries rep g repeat 
	          for inEdge in ndinfo.inEdges repeat 
		     yield inEdge } ];
  }
   
  outEdges(g: %, n: T): List Edge T == rep(g).n.outEdges;
  inEdges(g: %, n: T):  List Edge T == rep(g).n.inEdges;

  addLink!(g: %, start: T, end: T): () == {
       import from String;
       import from Edge T;
       import from Partial NdInfo;
       import from TextWriter;

stdout << "Adding: " << start << " -> " << end << newline; 
       if not node?(start, g) then { 
	  stderr << "Missing start node: " << start << " -> " << end << newline; 
	  never;
       }
       if not node?(end, g) then {
	  stderr << "Missing end node: " << start << " -> " << end << newline; 
          never;
       }
       startNd := (rep g).start;
       endNd   := (rep g).end;
       e := edge(start, end);
       endNd.inEdges    := cons(e, endNd.inEdges);
       startNd.outEdges := cons(e, startNd.outEdges);
   }

  -- topologically sort the given graph
  sortedNodes(g: %): Set T == {
      l: List T := [];
      visitTbl: Set T := [];
      for n in nodes g repeat {
	  (l, visitTbl) := sortVisit!(g, n, l, visitTbl);
      }
      return [id for id in l];
  }
 
  local sortVisit!(g: %, n: T, l: List T, tbl: Set T): (List T, Set T)
  == {
     import from List Edge T;

     if member?(n, tbl) then return (l, tbl);
     tbl := union!(tbl, n);

     for outEdge in outEdges(g, n) repeat {
         import from Edge T;
	 (l, tbl) := sortVisit!(g, p1 outEdge, l, tbl);
     }
     return (cons(n, l), tbl);
  }

  ancestors(g: %, n: T): Set T == {
     visitTbl: Set T := [];
     visitTbl := ancVisit!(g, n, visitTbl);
     return visitTbl;
  }

  local ancVisit!(g: %, n: T, tbl: Set T): Set T
  == {
     import from List Edge T;

     if member?(n, tbl) then return tbl;
     tbl := union!(tbl, n);

     for inEdge in inEdges(g, n) repeat {
         import from Edge T;
	 tbl := ancVisit!(g, p0 inEdge, tbl);
     }
     return tbl;
  }

   if T has OutputType then {
     (o: TextWriter) << (g: %): TextWriter == { 
      import from Set T;
      import from Edge T;
       import from String; 
       import from T;
       o << "[G: ";
       for node in sortedNodes(g) repeat {
	   o << node;
--	   sep: String := "";
--	   for e in inEdges(g, node) repeat {
--	     o << sep << p0(e);
--	     sep := ", ";
--	   }
--	   o << ")";
	   o << ", ";
       }      
       o << "]";
     }
   }
}


GraphAlgorithms(T: Join(HashType, OutputType), G: (T: Join(HashType, OutputType)) -> B__GraphCategory T): with {
   connectedComponents: G T-> Set Set T;
   connectedGraph: G T-> B__Graph Set T;
} == add {
  import from String;

  -- components that are strongly connected (ie. path from every elt to every
  -- other)
  connectedComponents(g: G T): Set Set T == {
	  import from List Edge T;
	  import from Edge T;
	  local components: Set Set T := [];
	  local fwdTbl: Set T := [];
	  local backTbl: Set T := [];
	  local visitOrder: List T := [];

          fwdVisit(node: T): ()  == {
	      free fwdTbl := union!(fwdTbl, node);
	      for outEdge in outEdges(g, node) repeat {
	          nxtNode := p1 outEdge;
		  if (not member?(nxtNode, fwdTbl)) then fwdVisit(nxtNode);
	      }
	      free visitOrder := cons(node, visitOrder);
	  }

          backVisit(node: T): Set T  == {
	      free backTbl := union!(backTbl, node);
	      mbrs: Set T := [node];
	      for inEdge in inEdges(g, node) repeat {
	          nxtNode := p0 inEdge;
		  if (not member?(nxtNode, backTbl)) 
		     then mbrs := union!(mbrs, backVisit(nxtNode));
	      }
	      mbrs
	  }

	  -- /* First, do nodes forwards */
	  for node in nodes(g) repeat {
	      if (not member?(node, fwdTbl)) then 
		 fwdVisit(node);
	  }
	  
	  -- /* And backwards */
	  for node in visitOrder repeat {
	      if (not member?(node, backTbl)) then {
		 m := backVisit(node);
		 components := union!(components, m);
	      }
	  }
	  return components;
   }

   connectedGraph(g: G T): B__Graph Set T == {
      import from G T;
      import from G Set T;
      import from Set Set T;
      import from Set T;
      import from List Edge T;
      import from Edge T;

      nodesC := connectedComponents g;
      gC := graph(c for c in nodesC);

      t: HashTable(T, Set T) := table();
      for component in nodesC repeat 
	  for node in component repeat
	      t.node := component;

      for component in nodesC repeat {
	  newEdges: Set Set T := [];
          for node in component repeat {
	      edges := outEdges(g, node);
	      for e in edges repeat 
	         if not member?(p1 e, component) then newEdges := union!(newEdges, t(p1 e));
	  }
	  for e in newEdges repeat addLink!(gC, component, e);
      }

      return gC;
   }

}

-----
#if LooksLikeThisDoesntWork
#include "aldor"

SetLike(T: PrimitiveType): Category == PrimitiveType with { first: % -> T }

define GraphCat(T: PrimitiveType): Category == with {
      NodeSet: () -> SetLike T;
      nodes: % -> NodeSet();
} 

Graph(T: PrimitiveType): GraphCat T == add {
  NodeSet(): SetLike T with == add {  
         first(x: %): T == never;
         (a: %) = (b: %): Boolean == never;
  }

  nodes(g: %): NodeSet() == never;
}
#endif